# -*- coding:utf-8 -*-
# @Time: 2022/12/11 11:46
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: typeutils.py

from typing import Union

REAL_NUM_TYPE = Union[float, int]
NUMERICAL_TYPE = Union[REAL_NUM_TYPE, complex]
